import pandas as pd
import pathlib
import re

class DataLoader:
    """
    Used to load data from excel, csv or text file and add to pandas object

    All data in dataframe is stored as string. User should convert to specific type if needed.

    QuickStart:
    input_file_path = './ExampleData.xlsx' 
    x = DataLoader(input_file_path)
    df = x.LoadFile()
    print(df)

    Input Paramters:
    Load data into pandas dataframe. This data will be processed for creating transact sql files.

    input_file_path: file of input excel, csv or text file. Must have correct delimiter
    delimiter: deliminter for text or excel file
    formatFunction: Preprocessing done to excel or text files. All columns will be a string type.
    hasHeader: has header rows
    skipRows: list or row numbers to skip. starting from top. list of rows i.e. [1,2]
    skipFooter: list of row number to skip starting from botton. list of rows i.e. [1,2]
    df: pandas object that stores data from excel/text file
    alias: object with df column and name of column to rename it to. i.e. {"dept": "Department"} if dept is name
        of column in data frame will be replace with Department. Columns are renamed before custom format function is
        run. so make sure to use values in alias for columns if using format function
    """
    def __init__(self,input_file_path, delimiter = ',', formatFunction = None, hasHeader=True, skipHeaderRows=None, skipFooterRows=None, stripDateTime=True,
        alias=None  ):
        self.input_file_path = input_file_path
        self.delimiter = delimiter
        self.formatFunction = formatFunction #input is the dataframe for self.df
        self.hasHeader = hasHeader
        self.skipHeaderRows = skipHeaderRows
        self.skipFooterRows = skipFooterRows
        self.stripDateTime = stripDateTime #If true or list
        self.alias = alias
        self.df = None

    def LoadFile(self):
        """
        Main function to run and return data file in pandas format
        """
        self.RunDataPull()
        return self.df

    def RunFormatFunction(self):
        """
        Run user defined format function to add or modify rows pulled from text file or excel. The formatFunction passed
        should take a dataframe as an object and do any processing required before returning back to user.

        All changes to the dataframe should be saved in the input object dataframe. So if using pandas functions
        used the inplace=True flag.

        This is the final function ran when using LoadFile.
        """
        if self.formatFunction is None:
            return
        df = self.df
        self.formatFunction(df)

    def RunDataPull(self):
        """
        Load and process data from excel, csv or text file.
        """
        self.LoadDataFrame()
        self.FillNa()
        self.ReplaceSingleQuote()
        self.StripTimeFromDateTime()
        self.RenameColumns()
        self.RunFormatFunction()

    def RenameColumns(self):
        """
        self.alias is a dictionary. Key value pairs are the original column names in pandas dataframe
        and new column new to replace it with
        """
        if self.alias is None:
            return
        self.df.rename(columns=self.alias, inplace=True)

    def LoadDataFrame(self):
        """
        pulledDataFrame = None
        fext = pathlib.Path(self.input_file_path).suffix

        if fext in ['xls', 'xlsx']:
            pulledDataFrame = pd.read_excel(self.input_file_path)
        else:
            pulledDataFrame = pd.read_csv(self.input_file_path, delimiter = self.delimiter)        
        """
        fext = pathlib.Path(self.input_file_path).suffix
        params = self.LoadParams(fext)
        self.df = self.__LoadDataFrame__(params, fext)

    def LoadParams(self, fext):
        """
        Converts all columns to string. Add additional parameters for reading
        file by pandas
        """
        params = {'dtype': str, 'parse_dates': False}
        if not self.hasHeader:
            params['header'] = None
        if self.skipHeaderRows != None:
            params['skiprows'] = self.skipHeaderRows
        if self.skipFooterRows != None:
            params['skipfooter'] = self.skipHeaderRows
        if self.delimiter != '' and fext not in ['.xls', '.xlsx']:
            params['delimiter'] = self.delimiter
        return params

    def __LoadDataFrame__(self, params, fext):
        #read excel or text/csv based on file extension
        fext = pathlib.Path(self.input_file_path).suffix
        if fext in ['.xls', '.xlsx']:
            df = pd.read_excel(self.input_file_path , **params)
            for columnName, columnData in df.iteritems():
                df[columnName] = df[columnName].astype(str)
            return df
        else:
            df = pd.read_csv(self.input_file_path , **params)
            for columnName, columnData in df.iteritems():
                df[columnName] = df[columnName].astype(str)
            return df


    def FillNa(self):
        """
        Fill empty rows with empty string
        """
        self.df.fillna('', inplace=True)

    def ReplaceSingleQuote(self):
        """
        Checks to see if any value in the excel/text file contains a single quote. Single quotes need to be escaped
        for by '' for including in sql file
        """
        #if xlsx
        fext = pathlib.Path(self.input_file_path).suffix
        for columnName, columnData in self.df.iteritems():
            self.df[columnName] = self.df[columnName].str.replace(u"’", u"'")

        keys = self.df.keys()
        for index, row in self.df.iterrows():
            for key in keys:
                value = row[key]
                if "'" in value:
                    self.df.at[index,key] = value.replace("'","''")

    def StripTimeFromDateTime(self):
        """
        Checks to see if any value in the excel/text file contains a single quote. Single quotes need to be escaped
        for by '' for including in sql file

        Assumes trailing zeros should be ignored. Can specify columns if need to keep
        """
        if self.stripDateTime == False:
            return
        elif isinstance(self.stripDateTime, list):
            for columnName in self.stripDateTime:
                self.df[columnName] = self.df[columnName].str.replace('00:00:00$', '', regex=True)        

        for columnName, columnData in self.df.iteritems():
            self.df[columnName] = self.df[columnName].str.replace('00:00:00$', '', regex=True)

if __name__ == '__main__':
    input_file_path = './ExampleData.xlsx' 
    x = DataLoader(input_file_path, delimiter = ',', formatFunction = None, hasHeader=True, skipHeaderRows=None, skipFooterRows=None, stripDateTime=True  )
    df = x.LoadFile()
    x.StripTimeFromDateTime()
    df['concat'] = df['Billing'] + df['Collections']
    print(df)

    input_file_path = './ExampleData.csv' 
    x = DataLoader(input_file_path, delimiter = ',', formatFunction = None, hasHeader=True, skipHeaderRows=None, skipFooterRows=None)#, stripDateTime=False)
    df = x.LoadFile()
    x.StripTimeFromDateTime()
    print(df)


"""
main function for parsing .xls, .xlsx and .csv files and sending data to a tsql server.
.sql files can be ran direclty.
"""
#pip install tsql_xlsx
import argparse
from .data_loader import DataLoader
from .tsql_connector import TsqlConnect
from .sql_ddl import SqlDDL
import argparse
import os
import pyodbc
import pathlib
import re

#argparse:
def mkdir (file_path):
    try:
        os.makedirs(file_path)
    except Exception as e:
        # print(e)
        pass

dtypex = ['bigint', 'numeric', 'smallint', 'int', 'float', 'real', 'date', 'datetime', 'datetime2', 'time',
    'varchar(250)', 'varchar(500)', 'varchar(100)', 'varchar(1500)', 'varchar(2000)' ]


class RunMain:

    def __init__(self):
        self.dargs = None

    def Run(self):
        #Step 1. Fill values
        dargs = self.dargs
        server   = dargs['s']
        database = dargs['d']
        username = dargs['u']
        password = dargs['p']
        schema_name = dargs['schema_name']
        table_name  = dargs['table_name']

        ctype_map = { 'ci': 'create_insert', 'c': 'create', 'i': 'insert',
            'u': 'update', 'd':'delete'}


        ctype       = ctype_map [ dargs['ctype'] ]
        use_trusted = dargs['use_trusted']
        username = dargs['u']
        password = dargs['p']
        pyodbc_driver = dargs['pyodbc_driver']
        in_file_name   = dargs['f']
        out_file_name  = dargs['o']
        odir = dargs['odir']
        verbose = dargs['v']
        alias   = dargs['alias']
        delim   = dargs['delim']
        as_null = dargs['null']

        column_types   = dargs['column_types']
        where_fields   = dargs['where_fields']
        set_fields     = dargs['set_fields']

        stripDateTime  = dargs['std']
        batchSize      = dargs['bsize']
        qstring        = dargs['q']

        # Run(self,sqlFileName, is_file=True, verbose=True, output_file="")

        if dargs['find_params']: #argument for first hit
            tsqlx = TsqlConnect(server,database,  username = username,password =password, 
                useTrusted=True, pyodbcDriver = "", store_df=False, sqlCmdSep=delim)
            tsqlx.FindCorrectConnection(first_hit=False)
            return

        if dargs['find_params_1']: #argument for first hit
            tsqlx = TsqlConnect(server,database,  username = username,password =password, 
                useTrusted=True, pyodbcDriver ='', store_df=False, sqlCmdSep=delim)
            tsqlx.FindCorrectConnection(first_hit=True)
            return

        if dargs['ld']: #argument for first hit
            print(pyodbc.drivers())
            return

        if dargs['pbat']:
            in_path = pathlib.Path(__file__).parent.resolve()
            fpath   = os.path.join(in_path, 'run_edw.bat')
            f = open(fpath)
            s = f.read()
            print(s)
            f.close()
            return

        if dargs['ppy']:
            in_path = pathlib.Path(__file__).parent.resolve()
            fpath   = os.path.join(in_path, 'example_py.txt')
            f = open(fpath)
            s = f.read()
            print(s)
            f.close()
            return


        #run raw query string
        if not dargs['q'] is None:
            sql_string = dargs['q']
            tsqlx = TsqlConnect(server,database,  username = username,password =password, 
            useTrusted=use_trusted, pyodbcDriver =pyodbc_driver)
            tsqlx.Run(sql_string, is_file=False,verbose=verbose,output_file=out_file_name)
            return

        #convert xls, xlsx and csv to sql files and run
        fext = pathlib.Path(in_file_name).suffix
        if fext in ['.xls', '.xlsx', '.csv']:
            input_file_path = in_file_name
            x = DataLoader(input_file_path, delimiter = delim, stripDateTime=stripDateTime, alias=alias  )
            df = x.LoadFile()
            #sql script file creation:
            table_name  = table_name
            crud_type = ctype
            sql_file = os.path.join(odir, table_name)
            csqlx  = SqlDDL(df, schema_name, table_name, crud_type,  sql_file, batchSize = batchSize,
                set_fields = set_fields, where_fields = where_fields, 
                column_types=column_types, as_null=as_null)
            csqlx.CreateFiles()

            if dargs['sexe']:
                return


            for sql_file in csqlx.sql_output_file_list:
                tsqlx = TsqlConnect(server,database,  username = username,password =password, 
                    useTrusted=use_trusted, pyodbcDriver =pyodbc_driver)
                tname = pathlib.Path(sql_file).stem
                print(sql_file)
                of = ""
                if out_file_name != "":
                    of = out_file_name + '_'+tname.replace(" ","_")
                tsqlx.Run(sql_file, is_file=True,verbose=verbose, output_file=of)
            return

        #run sql file
        if fext == '.sql':
            sql_file = in_file_name
            tsqlx = TsqlConnect(server,database,  username = username,password =password, 
            useTrusted=use_trusted, pyodbcDriver =pyodbc_driver)
            tsqlx.Run(sql_file, is_file=True,verbose=verbose,output_file=out_file_name)
            return

    #simple script template
    def InitializeParser(self):
        parser = argparse.ArgumentParser(description='Process some integers.')
        # parser.add_argument('integers', metavar='N', type=int, nargs='+',
        #                     help='an integer for the accumulator')
        findp_h =""" this flag makes serveral attempts to connect to a tsql server with different conneciton parameters. The first
        successful hit is generaly the best configuration to use. Please specify username and password. This is used as a last
        resort for connections that wont allow for the use_trusted flag. Python require the sql server drivers to be installed into
        the operating system. Version 13 seems to work consistently.

        SQL Drivers can be found at: These are the executables for connecting to sql server.
        https://docs.microsoft.com/en-us/sql/connect/odbc/windows/release-notes-odbc-sql-server-windows?view=sql-server-ver15#previous-releases
        """
        parser.add_argument('--find_params', help= findp_h, action='store_true')
        parser.add_argument('--find_params_1', help= "same as --find_params but stops on first succesful hit", action='store_true')
        parser.add_argument('--ld', help= "list all pyodbc drivers available", action='store_true')
        parser.add_argument('--pbat', help= "dislpay example.bat file for running program as windows executable", action='store_true')
        parser.add_argument('--ppy', help= "dislpay example.py file for running program as python module", action='store_true')


        parser.add_argument('-s', help= 'name of the server', default= "")
        parser.add_argument('-d', help= 'name of the database.', default= "")
        parser.add_argument('-schema_name', help= 'name of the schema', default= "")

        parser.add_argument('-table_name', help= 'Name of table. Default is the name of the input file with extension removed if its an .xls, .xlsx or .csv')
        ctype_h = """ name of the sql operation. default is ci
        ci: create/drop table followed by a series of insert tables into [schema].[table]
        c: create table only using columns from file
        i: series of insert tables into [schema].[table]
        u: series of update tables into [schema].[table] Requires where_fields and set_fields to be specified.
            this is a list of columns in the file that will be used in the where and set statements
        d: series of delete tables into [schema].[table] Requires where_fields to be specified.
            this is a list of columns in the file that will be used in the where statement
        """
        parser.add_argument('-ctype', help= ctype_h, choices=['ci','c','i','d','u'], default= "ci")


        parser.add_argument('-use_trusted', help= 'use credentials from vpn. this option does not require username or password', action='store_true' )
        parser.add_argument('-u', help= 'username i.e. bgovindara@iuhealth.org only required if use_trusted is false', default="")
        parser.add_argument('-p', help= 'password only required if use_trusted is false', default="")

        parser.add_argument('-pyodbc_driver', help= """name of the pyodbc drive to use i.e. "SQL SERVER 13" If no driver is specified, 
            will attempt to run sqlcmd direclty. This is not the preferred method""", default="")


        parser.add_argument('-f', help= 'Path to excel or csv file. Header must be present in file. if extension is .sql will run commands in the file directly')
        parser.add_argument('-sexe', help= 'Optional. Skip execution of sql files generated from .xls, .xlsx or .csv files', action='store_true', default=False)


        parser.add_argument('-q', help= 'Raw query string. Make sure to wrap in quotes i.e. -q "Select 1 from table;"')

        o_help = """
            output file path. if pyodbc_driver is specified valid extenisons are .xls, .xlsx or .csv. i.e. ./pathToFileToWrite/Example.xlsx
            if .csv the delimiter to use can be specified by using -delim ","  The default is a comma. Any single character can be used
            as a delimiter
        """

        parser.add_argument('-o', help= o_help, default="")

        parser.add_argument('-odir', help= 'output directory for writing sql files generated from .xls, .xlsx and .csv')

        parser.add_argument('-v', help= 'Run program in verbose mode.', action='store_true' )

        parser.add_argument('-alias', help= """list of old_name new_name pairs where old_name is name of column in file
            example .xlsx has two column department and speciatly. to rename use
            -alias department DEPARTMENTS specialty Specialties. Use alias name if specifing data types for columns.        
        """, nargs="+", default = None)

        parser.add_argument('-set_fields', help= """Required for update i.e. -ctype u. list of column names used to create set argument.
        i.e. -set_fields department speciatly will create set department = department_value, specialty=specialty_value in sql files.
        name of columns should be names of alias columns if used
        """, nargs="+", default = None)

        parser.add_argument('-where_fields', help= """Required for update and delete i.e. -ctype u or -ctype d. list of column names used 
        to create where argument.i.e. -where_fields department speciatly will create 
        where department = department_value, specialty=specialty_value in sql files.
        name of columns should be names of alias columns if used
        """, nargs="+", default = None)

        parser.add_argument('-null', help= """list of column_names when missing should be set to sql server null value. Default behavior assumes value
        is a string and replaces values with 'NULL' where Null is a string. 
        """, nargs="+", default = None)

        parser.add_argument('-bsize', help="Batch size. Max number of sql commands sent in one file. 7000 appears to be the max allowed", default = 7000, type = int)

        help_std = """Date columns in excel are automatically converted to date time. In order to skip this effect use the -std flag. 
        If you only want the behavior for certain columns use -std col_name_ col_name_2
        """

        parser.add_argument('-std', help=help_std, default=False, const=True, nargs='?')

        parser.add_argument('-delim', help= 'delimiter for csv files', default=",")

        self.AddDtypes(parser)
        args = parser.parse_args()
        dargs = vars(args)
        self.CheckAndParseArgumentConditions(dargs)
        self.dargs = dargs

    def CheckAndParseArgumentConditions(self, dargs):
        """
        This check all conditional arguments and also parse arguments for downstream functions
        """

        #default use_trusted if u == "" and p == ""
        if not dargs['use_trusted'] and dargs['u'] == "" and dargs['p'] == "":
            dargs['use_trusted'] = True

        dargs['column_types'] = None
        if dargs['find_params']:
            if dargs['u'] == "" or dargs['p'] == "":
                print("username and password should be added to params when checking for all possible configurations")
            return

        if dargs['find_params_1']:
            if dargs['u'] == "" or dargs['p'] == "":
                print("username and password should be added to params when checking for all possible configurations")
            return

        if dargs['ld']:
            return

        if dargs['pbat']:
            return

        if dargs['ppy']:
            return
        #print help
        if not dargs['find_params'] and not dargs['find_params_1'] and not dargs['ld'] and not dargs['pbat'] and not dargs['ppy'] and dargs['f'] is None and dargs['q'] is None:
            raise Exception('No arguments specified use -h for help')

        if dargs['s'] == "" and not dargs['sexe']:
            raise Exception('The name of the server must be defined')

        #get file path
        if dargs['f'] is None and dargs['q'] is None:
            raise Exception('-f /path/to/file must be specified. or -q with query string')


        #add something for -q
        if not dargs['f'] is None:
            fp = dargs['f']
            if not os.path.exists(fp):
                raise Exception ('the specified path was not found: ' + fp)

            fext = pathlib.Path(fp).suffix
            if fext in ['.xls', '.xlsx', '.csv']:
                if dargs['odir'] is None:
                    em = """The output directory for where sql files should be written from data generated from input files
                    needs to be specified with -odir /path/to/output_dir 
                    """
                    raise Exception (em)
                mkdir(dargs['odir'])
                op = dargs['odir']
                if not os.path.exists(op):
                    raise Exception ('attempted to created directory path specified but failed. Make sure its a valid path for a directory: ' + op)

                if dargs['schema_name'] == "":
                    raise Exception ('schema_name must be defined.')


                if dargs['table_name'] is None:
                    tname = pathlib.Path(fp).stem
                    dargs['table_name'] = re.sub('[^a-zA-Z0-9_]', '', tname.replace(" ","_") )

        elif dargs['q'] is None:
            raise Exception ('-q or -f was not used. Nothing to do')

        if not dargs['use_trusted']:
            if dargs['u'] == "" or dargs['p'] == "":
                raise Exception("username and password must both be specified. Otherwise dont use either. Default has use_trusted set to true")

        if not dargs['alias'] is None:
            if len(dargs['alias']) % 2 != 0:
                raise Exception("alias list had uneven amount of entries make sure each old_column has a matching new_column name.")
            if len(dargs['alias']) > 0:
                alias_map = {}
                ax = dargs['alias']
                for i in range(0,len(ax),2):
                    alias_map[ax[i]] = ax[i+1]
                dargs['alias'] = alias_map

        # dargs['column_types'] = None
        column_types = {}
        for dtx in dtypex:
            if dargs[dtx] is None:
                continue
            column_names = dargs[dtx]
            for cn in column_names:
                column_types[cn] = dtx
        if len(column_types) > 0:
            dargs['column_types'] = column_types

        if dargs['ctype'] in ['u', 'd']:
            cx = dargs['ctype']
            if dargs['where_fields'] is None:
                raise Exception("-ctype " +x + "requires where_fields to be specified.")
            if cx == 'u':
                if dargs['set_fields'] is None:
                    raise Exception("-ctype -u requires set_fields to be specified.")


    def AddDtypes(self, parser):
        """
        Creates a list of flags for specifing data conversion types for reading from file.
        """
        for dtx in dtypex:
            arg = '-' + dtx
            helpx = 'list of columns to set as type '+dtx +'. ex: '+arg + ' column_name1 column_name2'
            parser.add_argument(arg, help=helpx, nargs="+")


if __name__ == '__main__':
    x = RunMain()
    x.InitializeParser()
    x.Run()



import pandas as pd
import subprocess
import collections
import math
import pathlib
import os

class SqlDDL:
    """
    Used to create corresponding sql files for sql server processing. Uses column information from pandas dataframe to generate
    sql files

    Creates sql file to refresh data

    Quickstart:
    df   = pd.read_excel("./ExampleData.xlsx", dtype="string")
    csqlx  = SqlDDL(df, "test", "create",  "./test/create.sql", column_types={"id":"integer"})
    csqlx.CreateFiles()

    Parameters:
    df: pandas data frame object to use for create sql statements. All values should be set to string
        i.e. pd.read_excel("ExcelFile.xlsx",dtype="string")
    schema_name: name of schema/database
    table_name: Name of table to preform crud operations on
    column_types: dictionary {columnName: data_type} i.e. {"department": "text"} The columnName should be in the
        pandas dataframe. The data_type is a correpsonding sql data type. Column types are only needed when using
        create or create_insert crud_type. If column type is not specified text is used by default.


    crud_type: List of sql statements that can be created from dataframe. All require schema_name and table_name
        create: Drop/Create a new table. Optionally can specify column_types
        create_insert: Drop/Create a new table. Optionally can specify column_types. Also creates batch insert commands
            from data in the data frame.
        insert: creates batch insert commands from data in the data frame.
        update: creates batch update commands. Requires list of set_fields and list of where_fields to run
        delete: creates batch delete commands. Requires list of where_fields to run

    batchSize: 7000 How many crud statements per file.
    output_file_path: 

    set_fields: list of columns used to set new values on update crud operation. i.e. ["provider"] will create
        set [provider] = provider_value

    where_fields: list of columns used to specifiy query conditions for update/delete i.e. ["department","specialty"]
        will create: where [department] = department_value, [specialty] = specialty_value
    as_null: list of column names where an actual NULL value should be placed in sql file where appropriate. replaces
        df values with "none", "null", "", None with sql NULL

    sql_output_file_list: Stores list of absolute files paths to created files. Used to keep track of file order creation
        for batch file creation.
    """
    def __init__(self,df,schema_name  , table_name, crud_type,  output_file_path, batchSize = 7000, set_fields = None, where_fields = None, 
        column_types=None,  as_null = None ):
        self.df = df
        self.table_name = table_name
        self.crud_type = crud_type
        self.output_file_path = output_file_path
        self.batchSize = 7000
        self.set_fields = set_fields #should be list of columns to use as set value. should be name of column after alias
        self.where_fields = where_fields
        if column_types is None:
            column_types = {}
        self.column_types=column_types #key value pair. column name and type

        self.schema_name=schema_name
        #self.sqlFileList
        self.batch_list = None
        self.as_null = as_null #List of column names that should have NULL used in place of "NULL" pandas df null or ""
        self.sql_output_file_list = [] #stored in file creation order. Used for sending sql files to sql server in order.

    def CreateFiles(self):
        self.ValidateInputParameters()
        self.GetBatches()
        self.WriteFiles()

    def GetBatches(self):
        #Add loop based on index
        numLoops = math.ceil( len(self.df)/(1.0*self.batchSize) )
        batch_list = []
        print("Number of files to be created: "+str(numLoops))
        for i in range(0,numLoops):
            indexStart = i*self.batchSize
            indexEnd = (i+1)*self.batchSize
            ix = {'start_index': indexStart, 'stop_index': indexEnd}
            batch_list.append(ix)
        self.batch_list = batch_list

    def ValidateInputParameters(self):
        if self.crud_type not in ['create_insert', 'create', 'insert', 'update', 'delete']:
            pass
        if self.crud_type == 'update' and (self.set_fields is None or self.where_fields is None):
            raise Exception("set_fields and/or where_fields not defined for update")
        if self.crud_type == 'delete' and self.where_fields is None:
            raise Exception("where_fields not defined for delete")

        if self.batchSize <= 0:
            self.batchSize = 7000

    def WriteFiles(self):
        if self.crud_type == 'create_insert':
            self.WriteCreateInsertStatements()

        elif self.crud_type == 'create':
            self.WriteCreateStatement()

        elif self.crud_type == 'insert':
            self.WriteInsertStatements()

        elif self.crud_type == 'update':
            self.WriteUpdateStatements()

        elif self.crud_type == 'delete':
            self.WriteDeleteStatements()
        else:
            raise Exception("""No valid crud type select. if on command line use ci, c, i, u or d 
                which represent create_insert, create, insert, update and delete in the module""")

    #Write functions used for writing sql to file
    def WriteCreateInsertStatements(self):
        createStr = self.CreateDropStatement()
        insertStr = self.CreateInsertStatements()
        for i in range(0,len(self.batch_list)):
            bk = self.batch_list[i]
            insertSelectStr = self.CreateSelectStatements( bk['start_index'], bk['stop_index'])
            sqlString = None
            if i == 0:
                sqlString = '\n'.join( [createStr, insertStr, insertSelectStr] )
            else:
                sqlString = '\n'.join( [insertStr, insertSelectStr] )
            sqlFileName = self.CreateSqlFileNamePath(i)
            self.WriteSqlFileCommand(sqlFileName,sqlString)

    def WriteCreateStatement(self):
        createStr = self.CreateDropStatement()
        sqlFileName = self.CreateSqlFileNamePath(0)
        self.WriteSqlFileCommand(sqlFileName,createStr)

    def WriteDeleteStatements(self):
        for i in range(0,len(self.batch_list)):
            bk = self.batch_list[i]
            sqlString = self.CreateDeleteStatements( bk['start_index'], bk['stop_index'])
            sqlFileName = self.CreateSqlFileNamePath(i)
            self.WriteSqlFileCommand(sqlFileName,sqlString)
    
    def WriteInsertStatements(self):
        insertStr = self.CreateInsertStatements()
        for i in range(0,len(self.batch_list)):
            bk = self.batch_list[i]
            insertSelectStr = self.CreateSelectStatements( bk['start_index'], bk['stop_index'])
            sqlString = '\n'.join( [insertStr, insertSelectStr] )
            sqlFileName = self.CreateSqlFileNamePath(i)
            self.WriteSqlFileCommand(sqlFileName,sqlString)

    def WriteUpdateStatements(self):
        for i in range(0,len(self.batch_list)):
            bk = self.batch_list[i]
            sqlString = self.CreateUpdateStatements( bk['start_index'], bk['stop_index'])
            sqlFileName = self.CreateSqlFileNamePath(i)
            self.WriteSqlFileCommand(sqlFileName,sqlString)
    #End of Write functions used for writing sql to file

    #Create functions used to create sql strings
    def CreateUpdateStatements(self, indexStart, indexEnd):
        sql_list = []
        for index,row in self.df.iloc[indexStart:indexEnd].iterrows():
            set_string, where_string = self.UpdateSetWhereString(row)
            update_string = "UPDATE [{schema_name}].[{tableName}] set {set_string} where {where_string} ;".format(
                tableName=self.table_name, schema_name=self.schema_name, set_string=set_string, where_string=where_string)
            sql_list.append(update_string)
        return '\n'.join(sql_list)

    def CreateDeleteStatements(self, indexStart, indexEnd):
        sql_list = []
        for index,row in self.df.iloc[indexStart:indexEnd].iterrows():
            where_string = self.WhereString(row)
            delete_string = "DELETE FROM [{schema_name}].[{tableName}] where {where_string} ;".format(
                tableName=self.table_name, schema_name=self.schema_name, where_string=where_string)
            sql_list.append(delete_string)
        return '\n'.join(sql_list)

    def CreateInsertStatements(self):
        """
        The insert command is created with the defined columns entered in the order specified in
        self.column_types
        """
        insertString = "INSERT INTO [{schema_name}].[{tableName}] ".format(tableName=self.table_name, schema_name=self.schema_name)

        columnTypes = []
        for columnName in self.df.columns:
            columnTypes.append( "[{columnName}]".format(columnName=columnName) )

        columnTypesStr ='(' + ','.join(columnTypes)+')\n'
        insertString += columnTypesStr
        return insertString

    def CreateSelectStatements(self, indexStart, indexEnd):
        """
        Creates bulk insert string that allows for insert into sql server. This is where all the data from
        the excel file is transfered to the sql file. Union alls use older version syntax for sql server. So to do
        a bulk insert requires multiple select and Union Alls. 
        """
        columnTypes = []
        for columnName, dataType in self.df.iteritems():
            columnTypes.append( "{{{columnName}}}".format(columnName=columnName) )

        selectString ="SELECT " + ','.join(columnTypes)

        #begin adding data to sql sever string
        outList = [ ]
        for index,row in self.df.iloc[indexStart:indexEnd].iterrows():
            #string wrapping and unrapping here?
            rowx= self.StringifyPandasRow(row)
            x = selectString.format(**rowx)
            outList.append(x)

        outStr = "\nUNION ALL\n".join(outList) +';'
        return outStr


    def UpdateSetWhereString(self,row):
        """
        Creates where and set statements for sql string
        """
        rowx= self.StringifyPandasRow(row)
        set_list = []
        for sx in self.set_fields:
            set_list.append( "[{column_name}] = {value}".format(column_name=sx, value=rowx[sx])  )
        set_str = ', '.join(set_list)
        where_str = self.WhereString(row)
        return set_str, where_str

    def WhereString(self,row):
        """
        Creates where statements
        """
        where_list = []
        rowx= self.StringifyPandasRow(row)
        for sx in self.where_fields:
            where_list.append( "[{column_name}] = {value}".format(column_name=sx, value=rowx[sx])  )
        return ' and '.join(where_list)

    def CreateDropStatement(self):
        """
        Creates the conditional drop if and create table statemetns. Uses the column names and data types in self.columnFields to 
        define the table to be created
        """

        createTableString = """IF OBJECT_ID('[{schema_name}].[{tableName}]','U') IS NOT NULL DROP TABLE [{schema_name}].[{tableName}] ;
        CREATE TABLE [{schema_name}].[{tableName}] 
        ( """.format(tableName=self.table_name, schema_name=self.schema_name)
        columnTypes = []
        for columnName in self.df.columns:
            dataType = "varchar(2000)"
            if columnName in self.column_types:
                dataType = self.column_types[columnName]
            columnTypes.append( "[{columnName}] {dataType}".format(columnName=columnName, dataType=dataType) )

        columnTypesStr = ','.join(columnTypes)+' );\n'
        createTableString += columnTypesStr
        return createTableString
    #Create functions used to create sql strings


    def StringifyPandasRow(self, row):
        """
        Parse values for null statements and makes strings of all values. single quotes
        wrap all values.
        """
        rowx = {}
        as_null = self.as_null
        if as_null is None:
            as_null = []

        if not self.column_types is None:
            for key, value in self.column_types.items():
                if value in  ['bigint', 'numeric', 'smallint', 'int', 'float', 'real', 'date', 'datetime', 'datetime2', 'time']:
                    if key not in as_null:
                        as_null.append(key)

        for column_name in row.keys():
            value = row[column_name]
            # tx = self.IsNull(value)
            if column_name in as_null:
                if self.IsNull(value) :
                    rowx[column_name] = "NULL"
                else:
                    rowx[column_name] = "'"+str(value) +"'"
            else:
                rowx[column_name] = "'"+str(value) +"'"
        return rowx

    def IsNull(self, value):
        tmp_str = str(value) 
        if pd.isna(value) or pd.isnull(value) or value is None:
            return True
        elif tmp_str.lower() in ["null", "none", "", "nan", "nat"]:
            return True
        else:
            return False


    def CreateSqlFileNamePath(self, idx_num, ignore_index=False):
        """
        Used to rename output file if necessary. And add indexes.
        """
        fext = pathlib.Path(self.output_file_path).suffix
        new_ext = ''
        if not ignore_index:
            new_ext =  "_"+str(idx_num)
        if fext == "":
            new_ext += '.sql'
        else:
            new_ext += fext
            return self.output_file_path.split(fext)[0] + new_ext
        return self.output_file_path + new_ext

    def WriteSqlFileCommand(self,sqlFileName,sqlString):
        """
        Combines all the sql file strings together into one string. Writes the
        string to the specified sql file.
        """
        self.sql_output_file_list.append( os.path.abspath( sqlFileName) )
        fout = open(sqlFileName,'w')
        fout.write(sqlString)
        fout.close()

if __name__ == '__main__':
    df   = pd.read_excel("./ExampleData.xlsx") #, dtype="string")
    for columnName, columnData in df.iteritems():
        df[columnName] = df[columnName].astype(str)


    as_null = ['dates']
    column_types={"id":"int"}

    csqlx  = SqlDDL(df,"schema_name", "test", "create_insert",  "./test/create.sql",  as_null = as_null, column_types = column_types)
    csqlx.CreateFiles()

    # column_types={"id":"integer"},

    # cisqlx = SqlDDL(df,"schema_name", "test", "create_insert",  "./test/create_insert.sql", column_types={"id":"varchar(255)", "dates": "date" })
    # cisqlx.CreateFiles()

    isqlx = SqlDDL(df,"schema_name", "test", "insert",  "./test/insert.sql", column_types={"id":"int", "dates": "date" })
    isqlx.CreateFiles()
    # usqlx = SqlDDL(df,"schema_name", "test", "update",  "./test/update.sql", column_types={"id":"varchar(255)", "dates": "date" }, set_fields=["department"], where_fields=["id"])
    # usqlx.CreateFiles()
    # dsqlx = SqlDDL(df,"schema_name", "test", "delete",  "./test/delete.sql", column_types={"id":"varchar(255)", "dates": "date" }, where_fields=["id"])
    # dsqlx.CreateFiles()


#Parameters
import subprocess
import os
import pyodbc
import pandas as pd
import pathlib


try:
    from .data_loader import DataLoader
except ImportError:
    print("DataLoader failed to load. Export dataframes from sqlcmd will not work")

class TsqlConnect:
    """
    SQL Drivers can be found at: These are the executables for connecting to sql server.
    https://docs.microsoft.com/en-us/sql/connect/odbc/windows/release-notes-odbc-sql-server-windows?view=sql-server-ver15#previous-releases

    Module used to connect to tsql and run sql files either as a string or a list of files.

    Connection information in the TsqlConnection class comments. To find the connection parameters run:
    server = ''
    database =''
    username = ''
    password = ""

    tsqlx = TsqlConnect(server,database,  username = username,password =password, 
        useTrusted=True, pyodbcDriver ='', store_df=False, sqlCmdSep='^')
    tsqlx.FindCorrectConnection(first_hit=False)

    Use Trusted is the preferred option. This does not require the username and password to be stored. But sometimes
    the local computer configurations dont work. There are many possible drivers that can connect to the tsql server. 

    if useTrusted true params:
        useTrusted = True (preferred method. )
        server = ''
        database =''        
    else:
        username = ""
        password = ""
        server = ''
        database =''
 
    pyodbcDriver = if empty string will use the sqlcmd command line option. If a string specifying the driver.
        Will use the pyodbc driver with the the specified driver string. Example
        pyodbcDriver = "ODBC Driver 13 for SQL Server"
        or another driver.

    error_string stores all errors from running sql commands.

    sqlCmdSep: the delimiter used when running sqlCmd. Also the delimiter used if writing to text or csv file.

    store_df: Default false. Stores query output into dataframe. Mainly for select statements
    output_df: List of dfs where dataframes are store after every query execution.

    """
    def __init__(self,server,database, username = '',password ='', 
        useTrusted=True, pyodbcDriver ='', store_df=False, sqlCmdSep='^'):

        self.server = server
        self.database = database
        self.username  = username
        self.password  = password

        self.useSqlCmd = pyodbcDriver == ''
        self.pyodbcDriver = pyodbcDriver
        self.useTrusted = useTrusted
        self.error_string = ''
        #supplemental values?
        self.sqlCmdSep = sqlCmdSep #deliminter used when writing results to file from sqlCmd. Also used as output separator for text files?
        self.store_df = store_df #used to determine if query results will be saved as pandas dataframe:
        #For using sqlcmd the output file must be specified when running query
        #Need output file for sqlcmd and output file for .xlsx
        self.output_dfs = []

    def RunBatch(self, sqlFileList, sqlOutputFileList=None):
        """
        sqlFileList: list of file paths to sql input scripts
        sqlOutputFileList: optional list of full paths to send each query. The list path should match the 
            input paths in sqlFileList 1:1.
        """
        #change to current directory
        if self.InvalidInputValues():
            return
        if not sqlOutputFileList is None:
            if len(sqlOutputFileList) != len(sqlFileList):
                raise Exception("The list for sqlFileList and sqlOutputFileList must be the same")

        for i in range(0,len(self.sqlFileList) ):
            output_file = ""
            sqlFileName = self.sqlFileList[i]
            if not sqlOutputFileList is None:
                output_file = sqlOutputFileList[i]
            self.Run(sqlFileName, is_file=True, verbose=True, output_file=output_file)

    def Run(self,sqlFileName, is_file=True, verbose=True, output_file=""):
        #change to current directory
        if self.InvalidInputValues():
            return

        if self.useSqlCmd:
            self.RunSqlCmd(sqlFileName, is_file=is_file, verbose=verbose, output_file = output_file)
        else:
            self.RunPyodb(sqlFileName, is_file=is_file, verbose=verbose, output_file = output_file)

    def RunQuery(self):
        #Used to run a select statement and return results
        pass

    def InvalidInputValues(self):
        if self.server == "":
            print("No server specified")
            return True
        if self.database == "":
            print("No server specified")
            return True
        if not self.useTrusted:
            if self.username == "" or self.password == "":
                print("username or password is not defined.")
                return True
        else:
            if self.username != "" or self.password != "":
                print("username and password is not required when running in trusted mode")
                return True
        return self.InvalidPyodbcValues()

    def InvalidPyodbcValues(self):
        """
        Checks if pyodbcDriver provided is valid
        """
        if not self.useSqlCmd:
            if self.pyodbcDriver == '':
                print("sqlDriver is not defined. Please select the proper drivers from list below")
                print( pyodbc.drivers() )
                return True
            else:
                if self.pyodbcDriver not in pyodbc.drivers():
                    print(self.pyodbcDriver+" is not available. Please select correct driver from list below")
                    print(pyodbc.drivers())
                    return True
                else:
                    return False
        else:
            return False


    def RunSqlCmd(self,sqlFileName, is_file=True, verbose=True, output_file = ""):
        """
        #executes the sqlcmd command
        """
        sqlCmd = self.CreateSqlServerCmd(sqlFileName,is_file=is_file, output_file = output_file)
        if verbose:
            print(sqlCmd)
            print('ExceutingCommand: '+ sqlFileName)        
        # subprocess.run(sqlCmd)
        x = subprocess.run(sqlCmd, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
        if x.stdout:
            stdout = x.stdout.decode("utf-8")
            print(stdout)
            #For storing data
            if self.store_df or (output_file != "" and fext != ".csv"):
                sqlFileName = output_file
                fext = pathlib.Path(sqlFileName).suffix
                if fext == '.xls':
                    sqlFileName = sqlFileName.split('.xls')[0] +'.csv'
                elif fext == '.xlsx':
                    sqlFileName = sqlFileName.split('.xlsx')[0] +'.csv'

                dlx = DataLoader(output_file, delimiter = self.sqlCmdSep, hasHeader=True, skiprows=[1], skipfooter=[1]  )
                dfx = x.LoadFile()
                if output_file != "" and fext != ".csv":
                    self.WriteDfToFile(dfx,output_file)
                if self.store_df:
                    self.output_dfs.append(df)

        if x.stderr:
            print("SQL Cmd Error")
            stderr = x.stderr.decode("utf-8")
            self.error_string = stderr
            raise Exception(stderr)

    def CreateSqlServerCmd(self,sqlStrFileName, is_file=True, output_file = ""):
        """
        #sqlCmd with trusted Connection?
        #creates the sqlcmd command to be run. Adds the username, password, and sqlfile to command string
        """
        sqlFileName = sqlStrFileName
        #if .xlsx or .xls replace with .csv
        if is_file:
            fext = pathlib.Path(sqlFileName).suffix
            if fext == '.xls':
                sqlFileName = sqlFileName.split('.xls')[0] +'.csv'
            elif fext == '.xlsx':
                sqlFileName = sqlFileName.split('.xlsx')[0] +'.csv'
            sqlCmd = 'sqlcmd -S {server} -d {database} -U "{username}" -P "{password}" -s "{seperator}" -W -G -I ' \
                '-i {sqlFile}'.format(seperator=self.sqlCmdSep,  username = self.username, password = self.password, server=self.server,database=self.database, sqlFile = sqlFileName)
            if self.useTrusted:
                sqlCmd = 'sqlcmd -S {server} -d {database} -E -s "{seperator}" -W -G -I ' \
                    '-i {sqlFile}'.format(seperator=self.sqlCmdSep, server=self.server, database=self.database, sqlFile = sqlFileName)
        else:
            sqlCmd = 'sqlcmd -S {server} -d {database} -U "{username}" -P "{password}" -s "{seperator}" -W -G -I ' \
                '-Q "{sqlStr}"'.format(seperator=self.sqlCmdSep, username = self.username, password = self.password, server=self.server,database=self.database, sqlStr = sqlFileName)
            if self.useTrusted:
                sqlCmd = 'sqlcmd -S {server} -d {database} -E -s "{seperator}" -W -G -I ' \
                    '-Q "{sqlStr}"'.format(seperator=self.sqlCmdSep, server=self.server, database=self.database, sqlStr = sqlFileName)

        if output_file != "":
            sqlCmd += " -o " +output_file

        return sqlCmd

    def RunPyodb(self,sqlStrFileName, is_file=True, verbose=True, output_file = ""):
        """
        sqlStrFileName: Path to sql file or if is_file=False the sql statement to be executed
        """
        sqlFileName = sqlStrFileName
        # print(self.pyodbcDriver)
        sqlDriverX = '{'+self.pyodbcDriver+'}'
        sqlConnString = "DRIVER={driver};".format(driver=sqlDriverX) + "SERVER={server};".format(server=self.server)+"DATABASE={database};".format(database=self.database)+ "UID={username};".format(username=self.username)+"PWD={password};".format(password=self.password)+"Authentication=ActiveDirectoryPassword;"
        #run connection using python back end trusted driver
        if self.useTrusted:
            sqlConnString = "DRIVER={driver};".format(driver=sqlDriverX) + "SERVER={server};".format(server=self.server)+"DATABASE={database};".format(database=self.database)+ "Authentication=ActiveDirectoryIntegrated;"

        sqlString = ""
        if is_file:
            sqlFile = open(sqlFileName)
            sqlString = sqlFile.read()
            sqlFile.close()
        else:
            sqlString = sqlStrFileName
        if verbose:
            print("Executing: "+sqlFileName)
            print(sqlConnString)
        try:
            conn = pyodbc.connect(sqlConnString)
            conn.autocommit=True
            cursor = conn.cursor()
            cursor.execute(sqlString)
            if output_file != "" or self.store_df:
                dfx = self.CreateDfFromPyodbQueryResults(cursor)
                if output_file != "":
                    self.WriteDfToFile(dfx,output_file)
                if self.store_df:
                    self.output_dfs.append(dfx)

            if verbose:
                print(str(cursor.rowcount) + " rows affected")

            #output file or df
            conn.close()
        except pyodbc.Error as e:
            print("SQL Pyboc cmd Error")
            self.error_string = e
            raise Exception(e)

    def CreateDfFromPyodbQueryResults(self, cursor):
        """
        Convert pyodbc driver data to df.
        """
        col_headers = [ i[0] for i in cursor.description ]
        rows = [ list(i) for i in cursor.fetchall()] 
        df = pd.DataFrame(rows, columns=col_headers)
        return df      

    def WriteDfToFile(self, df, output_file_path):
        """
        Writes data frame to file
        """
        fext = pathlib.Path(output_file_path).suffix
        if fext in ['.xls', '.xlsx']:
            df.to_excel(output_file_path, index=False)
        else:
            df.to_csv(output_file_path, sep=self.sqlCmdSep, index=False)


    def PrintPyodbcDrivers(self):
        print( pyodbc.drivers() )

    def FindCorrectConnection(self,first_hit=True):
        """
        #used to find right combination
        #run select 1; in order of most secure to least secure connection.
        #return command options for correct hits
        #if first_hit = True return results of first succesful run.
        """
        #Try Trusted Pyobc
        self.__AllFieldsRequiredForTestConn__()
        sxp = []
        username = self.username
        password = self.password
        is_hit = False
        for driverx in pyodbc.drivers():
            if self.SkipDriver(driverx):
                continue
            px = {'pyodbcDriver': driverx, 'useTrusted': True}
            self.ChangeParams(px)
            try:
                self.RunPyodb("Select 1;", is_file=False)
                sxp.append(px)
                if self.error_string == "":
                    is_hit = True
                    print("Connection Succesful with parameters: ", px )
                    if first_hit:
                        return
            except:
                pass

        px = {'pyodbcDriver': '', 'useTrusted': True}
        self.ChangeParams(px)
        try:
            self.RunSqlCmd("Select 1;", is_file=False)
            if self.error_string == "":
                is_hit = True
                px = {'pyodbcDriver': "", 'useTrusted': True}
                sxp.append(px)
                print("Connection Succesful with parameters: ", px )
                if first_hit:
                    return
        except:
            pass


        for driverx in pyodbc.drivers():
            if self.SkipDriver(driverx):
                continue
            px = {'pyodbcDriver': driverx, 'useTrusted': False, 'username': username, 'password': password}
            self.ChangeParams(px)
            if self.InvalidInputValues():
                continue
            try:
                self.RunPyodb("Select 1;", is_file=False)
                is_hit = True
                sxp.append(px)
                if self.error_string == "":
                    print("Connection Succesful with parameters: ", px )
                    if first_hit:
                        return
            except:
                pass

        px = {'pyodbcDriver': "", 'useTrusted': False, 'username': username, 'password': password}
        self.ChangeParams(px)

        if not self.InvalidInputValues():
            try:
                self.RunSqlCmd("Select 1;", is_file=False)
                if self.error_string == "":
                    print("Connection Succesful with parameters: ", px )
                    sxp.append(px)
                    if first_hit:
                        return
            except:
                pass

        if not is_hit:
            print("No combination of connections were succesful")
        else:
            print("Some Succesful Combinations")
            for x in sxp:
                print(x)

    def ChangeParams(self, params):
        if 'pyodbcDriver' in params:
            self.pyodbcDriver = params['pyodbcDriver']

        if 'useTrusted' in params:
            self.useTrusted = params['useTrusted']

        if 'username' in params:
            self.username = params['username']

        if 'password' in params:
            self.password = params['password']

        self.error_string = ''

    def SkipDriver(self, driver_name):
        skip_list = ['Microsoft', 'PostgresSQL','Amazon'] 
        for dx in skip_list:
            if dx in driver_name:
                return True
        return False


    def __AllFieldsRequiredForTestConn__(self):
        if self.username  == "" or self.password  == "":
            print('username and password must be defined for test connections. will skip testing that requires username and password. Only test trusted')

if __name__ == '__main__':
    server = ''
    database =''
    username = ''
    password = ""

    tsqlx = TsqlConnect(server,database,  username = username,password =password, 
        useTrusted=True, pyodbcDriver ='', store_df=False, sqlCmdSep='^')
    tsqlx.FindCorrectConnection(first_hit=False)